/*
 * Copyright (c) 1999, 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * netio.{cc,hh} -- Read packets from OSKit netio, and pass along.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "fromnetio.hh"
#include "error.hh"
#include "packet.hh"
#include "confparse.hh"
#include "glue.hh"
#include <unistd.h>
#include "netio_guts.h"

inline void
FromNetIO::enq(Packet *p)
{
	assert(p);
	
	int next = next_i(_qtail);
	if (next != _qhead) {
		_q[_qtail] = p;
		_qtail = next;
	} else {
		p->kill();
		_qdrops++;
	}
}

inline Packet *
FromNetIO::deq()
{
	if (_qhead != _qtail) {
		Packet *p = _q[_qhead];
		assert(p);
		_qhead = next_i(_qhead);
		return p;
	} else
		return 0;
}

FromNetIO::FromNetIO()
  : _ifnum(-1), _qhead(0), _qtail(0), _max(100), _qdrops(0)
{
	add_output();
}

FromNetIO::~FromNetIO()
{
	if (_ifnum >= 0) {
		(void) oskit_destroy_fromnetio(_ifnum, _priv);

		if (_qdrops)
			click_chatter("FromNetIO::~FromNetIO: "
				      "%d packets were dropped due to "
				      "Q overflow\n", _qdrops);
	}
}

FromNetIO *
FromNetIO::clone() const
{
	return new FromNetIO;
}

int
FromNetIO::configure(const String &conf, ErrorHandler *errh)
{
	if (cp_va_parse(conf, this, errh,
			cpInteger, "interface number", &_ifnum,
			cpEnd) < 0)
		return -1;

	return 0;
}

int
FromNetIO::initialize(ErrorHandler *errh)
{
	int	err;

	err = oskit_create_fromnetio(this, _ifnum, &_priv);
	if (err != 0)
		return err;

	_q = new Packet *[_max + 1];
	if (_q == 0)
		return errh->error("FromNetIO::initialize: out of memory");

	_qhead = _qtail = 0;
	return 0;
}

int
FromNetIO::got_packet(oskit_bufio_t *b)
{
	/*
	 * Make a packet out of the bufio data. If this succeeds, the
	 * packet constructor will bump the bufio reference count.
	 */
	Packet* p = Packet::make(b);

	/*
	 * Add to the queue. The packet will be dropped if the queue fills.
	 */
	enq(p);
	if (!scheduled())
		schedule_tail();
	
	return 0;
}

void
FromNetIO::run_scheduled()
{
	Packet *p = deq();

	if (! qempty())
		schedule_tail();

	assert(p);
	output(0).push(p);
}

/*
 * This is the upcall from the netio code.
 */
extern "C" int
click_FromNetio_in(void *obj, oskit_bufio_t *b, int)
{
	FromNetIO	*netio = (FromNetIO *) obj;

	return netio->got_packet(b);
}

EXPORT_ELEMENT(FromNetIO)
