/*
 * timer.{cc,hh} -- portable timers. Linux kernel module uses Linux timers
 * Eddie Kohler
 *
 * Copyright (c) 1999 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. For more information, see the `COPYRIGHT' file in the source
 * distribution.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "timer.hh"
#include "element.hh"
#include "router.hh"

#ifdef CLICK_OSKIT
/*
 * The Timer callback. Call the actual routine.
 */
int
Timer::handler(struct oskit_iunknown *, void *arg)
{
	Timer	*Tp = (Timer *) arg;

	Tp->_hook(Tp->_data);

	return 0;
}

/*
 * Create a Timer based on OSKit clock timers.
 */
Timer::Timer(TimerHook hook, unsigned long thunk)
	: _clock(0), _timer(0), _hook(0), _data(0), _scheduled(0)
{
	oskit_listener_t	*l;
		
	/* Create the clock (only one, and its probably already created). */
	_clock = oskit_clock_init();

	/* Create a timer */
	oskit_clock_createtimer(_clock, &_timer);

	/* Create a listener */
	l = oskit_create_listener(Timer::handler, (void *) this);

	/* Set the listener */
	oskit_timer_setlistener(_timer, l);
	oskit_listener_release(l);

	_hook = hook;
	_data = thunk;
	_scheduled = 0;
}

/*
 * Create a timer for an element.
 */
Timer::Timer(Element *f)
	: _clock(0), _timer(0), _hook(0), _data(0), _scheduled(0)
{
	oskit_listener_t	*l;
		
	/* Create the clock (only one, and its probably already created). */
	_clock = oskit_clock_init();

	/* Create a timer */
	oskit_clock_createtimer(_clock, &_timer);

	/* Create a listener */
	l = oskit_create_listener(Timer::handler, (void *) this);

	/* Set the listener */
	oskit_timer_setlistener(_timer, l);
	oskit_listener_release(l);

	_hook = element_timer;
	_data = (unsigned long) f;
	_scheduled = 0;
}

/*
 * Destroy a Timer.
 */
Timer::~Timer()
{
	if (_timer)
		oskit_timer_release(_timer);
	if (_clock)
		oskit_clock_release(_clock);
}

/*
 * Unschedule a Timer.
 */
void
Timer::unschedule()
{
	if (scheduled()) {
		oskit_itimerspec_t	timespec = { {0, 0}, {0, 0} };

		oskit_timer_settime(_timer, 0, &timespec);
	}
}

/*
 * Schedule a timer.
 */
void
Timer::schedule_after_ms(int ms)
{
	oskit_itimerspec_t	timespec;

	timespec.it_value.tv_sec     = ms / 1000;
	timespec.it_value.tv_nsec    = (ms % 1000) * 1000000;
	timespec.it_interval.tv_sec  = 0;
	timespec.it_interval.tv_nsec = 0;

	_scheduled = 1;
	
	oskit_timer_settime(_timer, 0, &timespec);
}

void
Timer::element_timer(unsigned long thunk)
{
	Element *f = (Element *)thunk;
	f->schedule_tail(); // don't do anything - just put it on the work list
}

#else  /* CLICK_OSKIT */

void
Timer::element_timer(unsigned long thunk)
{
  Element *f = (Element *)thunk;
  f->schedule_tail(); // don't do anything - just put it on the work list
#ifdef __KERNEL__
#ifndef CLICK_POLLDEV
  // run work list
  f->router()->run_scheduled();
#endif
#endif
}

#ifndef __KERNEL__

static Timer timer_head(0, (unsigned long)0);

void
Timer::static_initialize()
{
  timer_head._prev = timer_head._next = 0;
}

void
Timer::schedule_after_ms(int ms)
{
  if (scheduled())
    unschedule();
  click_gettimeofday(&_expires);
  struct timeval interval;
  interval.tv_sec = ms / 1000;
  interval.tv_usec = (ms % 1000) * 1000;
  timeradd(&_expires, &interval, &_expires);
  Timer *prev = &timer_head;
  Timer *trav;
  for (trav = prev->_next;
       trav && timercmp(&_expires, &trav->_expires, >);
       prev = trav, trav = trav->_next)
    /* nada */;
  _prev = prev;
  _next = trav;
  _prev->_next = this;
  if (trav) trav->_prev = this;
}

void
Timer::unschedule()
{
  if (scheduled()) {
    _prev->_next = _next;
    if (_next) _next->_prev = _prev;
    _prev = _next = 0;
  }
}

void
Timer::run_timers()
{
  struct timeval now;
  click_gettimeofday(&now);
  Timer *next;
  while ((next = timer_head._next) && !timercmp(&next->_expires, &now, >)) {
    timer_head._next = next->_next;
    if (next->_next) next->_next->_prev = &timer_head;
    next->_next = next->_prev = 0;
    next->_hook(next->_thunk);
  }
}

// How long select() should wait until next timer expires.
int
Timer::get_next_delay(struct timeval *tv)
{
  Timer *next;
  if((next = timer_head._next) == 0){
    tv->tv_sec = 1000;
    tv->tv_usec = 0;
    return(0);
  } else {
    struct timeval now;
    click_gettimeofday(&now);
    if(timercmp(&next->_expires, &now, >)){
#ifdef CLICK_NODEOS
      /* Return absolute time for cond_timedwait */
      timeradd(&next->_expires, &now, tv);
#else
      timersub(&next->_expires, &now, tv);
#endif
    } else {
      tv->tv_sec = 0;
      tv->tv_usec = 0;
    }
    return(1);
  }  
}

#endif /* !__KERNEL__ */
#endif /* CLICK_OSKIT */
