/*
 * Copyright (c) 1999, 2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <errno.h>

#include "lexer.hh"
#include "router.hh"
#include "error.hh"
#include "timer.hh"
#include "straccum.hh"
#include "elements/nodeos/frominchan.hh"
#include "elements/nodeos/tooutchan.hh"

extern "C" {
#include <nodeos.h>
#include <packet.h>
#include <channel.h>
#include <thread.h>
#include <mempool.h>
}

void	click_driver_loop(void *arg);

/*
 * This is in the auto generated elements.cc file.
 */
extern void export_elements(Lexer *);

/*
 * These next two are duplicated from the nodeos api_channel.c.
 */
/*
 * Memory associated with a channel.
 * This includes not only the channel object itself, but any items referenced
 * by the channel object; i.e., non-scalar data that is passed across the API
 * by-reference.  Arguably, the latter should be part of the channel object
 * and we should just copyin directly to the appropriate fields here, or the
 * channel implementation could do the copyin calls, but we are trying to
 * preserve the tenuous separation between this ``entry layer'' and the
 * implementation.
 */
struct chandata {
	channel_t chan;		/* must be first */
	channel_proto_t proto;
	channel_addr_t addr;
};

/*
 * @deallocations: channel_t, channel_proto_t and channel_addr_t
 * returned to the chan's flow's mempool.
 */
static void
chan_cleanup(channel_t *chan)
{
	mempool_free(channel_mempool(chan), chan, sizeof(struct chandata));
}

/*
 * This interface differs slightly because I want to use a configuration
 * file instead of a string. Feel free to change it if ya don't like it.
 */
extern "C" an_chan_t
an_clickchan_create(an_flow_t flow, char *config_string,
		    an_chan_t inchan, an_chan_t outchan)
{
	int		err = 0;
	thread_t	*mythread;
	struct chandata *chandata = 0;
	flow_t		*iflow = 0;
	channel_t	*iinchan, *ioutchan;
	int		innodeos = 0;
	static int	initialized;

	if (! initialized) {
		/*
		 * XXX Anything static is bound to be some kind of
		 * problem with multiple router graphs in operation.
		 */
		String::static_initialize();
		ErrorHandler::static_initialize(
			new FileErrorHandler(stderr, ""));

		initialized = 1;
	}

	/*
	 * Leave these here. Its better if the C++ initializers above
	 * are called first, before declaring C++ variables.
	 */
	Element	        *elem;
	Lexer		*lex    = 0;
	Router		*router = 0;
	ErrorHandler	*errh   = ErrorHandler::default_handler();
	String		graphstr;
	String		filename;

	/*
	 * First off, we need to create the graph.
	 */
#ifdef  CLICK_GRAPH_FILE
	FILE		*fp;
	StringAccum	config_sa;

	if ((fp = fopen(config_file, "r")) == NULL) {
		errh->message("Could not open config file: %s:", config_file);
		err = errno;
		goto done;
	}

	/*
	 * So, not sure why this is like this. Must read the file first and
	 * put it into String format.
	 */
	while (!feof(fp)) {
		char	*x;
		int	r;

		if ((x = config_sa.reserve(512)) == NULL)
			errh->fatal("Out of memory while reading config file");
		else {
			r = fread(x, 1, 512, fp);
			config_sa.forward(r);
		}
	}
	fclose(fp);
	graphstr = config_sa.take_string;
	filename = config_file;
#else
	graphstr = config_string;
	filename = "inline";
#endif
	/*
	 * Do all the lexer goo (parse config file).
	 */
	lex = new Lexer(errh);
	export_elements(lex);
	lex->save_element_types();
  
	lex->reset(graphstr, filename);
	while (lex->ystatement())
		/* do nothing */;
  
	router = lex->create_router();
	lex->clear();

	/*
	 * Initialize the router graph.
	 */
	if (router->initialize(errh) < 0) {
		errh->message("Could not initialize the router");
		err = EINVAL;
		goto done;

	}
	else {
		errh->message("Router configuration structure:");
		router->print_structure(errh);
	}

	/*
	 * Okay, we have a router graph. Since that all worked okay, lets
	 * do the nodeos specific goo.
	 */
	NODEOS_ENTER;
	innodeos = 1;

	/*
	 * Validate arguments.
	 */
	iflow = NODEOS_INTERN_FLOW(flow);
	iinchan = NODEOS_INTERN_CHAN(inchan);
	ioutchan = NODEOS_INTERN_CHAN(outchan);
	if (iflow == 0 || iinchan == 0 || ioutchan == 0) {
		err = EINVAL;
		goto done;
	}

	/*
	 * Allocate memory for the cut channel.
	 */
	chandata = (struct chandata *)
		   mempool_alloc(iflow->mpool, sizeof *chandata);
	if (chandata == 0) {
		err = ENOMEM;
		goto done;
	}

	/*
	 * We need to hook up the endpoints of the graph (FromInChan
	 * and ToInchan) to the channel objects.
	 */
	elem = 0;
	for (int i = 0; i < router->nelements(); i++) {
		if (! strcmp(router->element(i)->class_name(), "ToOutChan")) {
			elem = router->element(i);
			break;
		}
	}
	if (elem == 0) {
		errh->message("Could not find the ToOutChan element");
		err = EINVAL;
		goto done;
	}
	((ToOutChan *)elem)->set_chan(ioutchan);

	elem = 0;
	for (int i = 0; i < router->nelements(); i++) {
		if (! strcmp(router->element(i)->class_name(), "FromInChan")) {
			elem = router->element(i);
			break;
		}
	}
	if (elem == 0) {
		errh->message("Could not find the FromInChan element");
		err = EINVAL;
		goto done;
	}
	((FromInChan *)elem)->set_chan(iinchan);
	
	/*
	 * Create the Nodeos cutchannel. 
	 */
	if ((err = cut_channel_create(&chandata->chan, iflow, 0,
				      iinchan, ioutchan, click_FromInChan_in,
				      chan_cleanup)) != 0) {
		errh->message("Could not create cut channel");
		goto done;
	}

	/*
	 * XXX: I am munging the argument to the channel receive
	 * function.  Instead of the channel, I'll pass in the Click
	 * FromInChan C++ Element object so we can (more easily) jump
	 * back into the C++ world.
	 */
	iinchan->tdata.in.farg = elem;

	/*
	 * Create a kernel thread to run the driver loop
	 */
	if ((err = thread_start_kernel(iflow->tpool, (channel_t *)0,
				       click_driver_loop, (void *) router,
				       &mythread)) != 0) {
		errh->message("Could not create the kernel thread");
		goto done;
	}

	/*
	 * I guess we are done.
	 */
 done:
	if (err) {
		if (chandata)
			mempool_free(iflow->mpool, chandata, sizeof *chandata);
		if (router) {
			delete router;
			delete lex;
		}
			
	}
	if (innodeos)
		NODEOS_LEAVE;
	
	return err ? 0 : NODEOS_EXTERN_CHAN(&chandata->chan);
}

void
click_driver_loop(void *arg)
{
	Router		*router  = (Router *) arg;

	while (router->driver())
		;
}

