/*
 * Copyright (c) 1999 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Click COM component.
 */

extern "C" {
#include <oskit/c/malloc.h>
#include "click_com.h"
}

#include "lexer.hh"
#include "router.hh"
#include "error.hh"
#include "timer.hh"
#include "straccum.hh"

/*
 * XXX Belongs elsewhere!
 */
const struct oskit_guid oskit_clicker_iid = OSKIT_CLICKER_IID;

/*
 * This is in the auto generated elements.cc file.
 */
extern void export_elements(Lexer *);

/*
 * The actual component. 
 */
struct clicker {
	oskit_clicker_t		ioi;		/* The COM interface */
	int			count;		/* Ref count */
	int			stop;		/* MAKE IT STOP! */
	Lexer		       *lex;
	Router		       *router;
};

static OSKIT_COMDECL
clickcon_query(oskit_clicker_t *c,
	       const oskit_iid_t *iid, void **out_ihandle) 
{
	struct clicker	 *clicker = (struct clicker *) c;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	if (memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
	    memcmp(iid, &oskit_clicker_iid, sizeof(*iid)) == 0) {
		*out_ihandle = &clicker->ioi;
		++clicker->count;
		
		return 0;
	}

	*out_ihandle = NULL;
	return OSKIT_E_NOINTERFACE;
}

static OSKIT_COMDECL_U
clickcon_addref(oskit_clicker_t *c)
{
	struct clicker	 *clicker = (struct clicker *) c;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	return ++clicker->count;
}

static OSKIT_COMDECL_U
clickcon_release(oskit_clicker_t *c)
{
	struct clicker	 *clicker = (struct clicker *) c;
	unsigned newcount;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	if ((newcount = --clicker->count) == 0) {
		delete clicker->router;
		delete clicker->lex;
		sfree(clicker, sizeof(*clicker));
	}
	return newcount;
}

static OSKIT_COMDECL
clickcon_config(oskit_clicker_t *c, char *config_file)
{
	struct clicker	*clicker = (struct clicker *) c;
	ErrorHandler	*errh    = ErrorHandler::default_handler();
	FILE		*fp;
	String		filename = config_file;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	if ((fp = fopen(config_file, "r")) == NULL)
		errh->fatal("Could not open config file: %s", config_file);

	/*
	 * So, not sure why this like this. Must read the file first and
	 * put it into String format.
	 */
	StringAccum	config_sa;
	while (!feof(fp)) {
		char	*x;
		int	r;

		if ((x = config_sa.reserve(512)) == NULL)
			errh->fatal("Out of memory");
		else {
			r = fread(x, 1, 512, fp);
			config_sa.forward(r);
		}
	}
	fclose(fp);

	Lexer *lex = new Lexer(errh);
	export_elements(lex);
	lex->save_element_types();
  
	lex->reset(config_sa.take_string(), filename);
	while (lex->ystatement())
		/* do nothing */;
  
	Router *router = lex->create_router();
	lex->clear();

	clicker->lex    = lex;
	clicker->router = router;

	return 0;
}

static OSKIT_COMDECL
clickcon_start(oskit_clicker_t *c)
{
	struct clicker	*clicker = (struct clicker *) c;
	ErrorHandler	*errh    = ErrorHandler::default_handler();
	Router  	*router;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	router = clicker->router;
	
	if (router->initialize(errh) >= 0) {
		errh->message("Router configuration structure:");
		router->print_structure(errh);

		while (router->driver() && !clicker->stop)
			;
	}

	return 0;
}

static OSKIT_COMDECL_V
clickcon_stop(oskit_clicker_t *c)
{
	struct clicker	*clicker = (struct clicker *) c;

	assert(clicker != NULL);
	assert(clicker->count != 0);

	clicker->stop = 1;
}

static struct oskit_clicker_ops clickcon_ops = {
	clickcon_query,
	clickcon_addref,
	clickcon_release,
	clickcon_config,
	clickcon_start,
	clickcon_stop,
};

/*
 * Create an instance of a click object. The object is pretty much useless
 * until it is configured, and then started.
 */
oskit_clicker_t *
oskit_clicker_create(void)
{
	struct clicker	*c;
	static int	initialized;

	if (! initialized) {
		String::static_initialize();
		ErrorHandler *errh = new FileErrorHandler(stderr, "");
		ErrorHandler::static_initialize(errh);

		initialized = 1;
	}

	c = (struct clicker *) smalloc(sizeof *c);
	if (c == 0)
		return 0;

	c->ioi.ops = &clickcon_ops;
	c->count   = 1;
	c->stop    = 0;
	
	return &c->ioi;
}

